@@ -3,40 +3,71 @@
 /// * Set the correct color format for the swapchain
 /// * Second renderpass to draw the gui
 use std::collections::VecDeque;
+use std::convert::TryInto;
 use std::sync::Arc;
 use std::time::Instant;
-use std::convert::TryInto;
 
+use bytemuck::{Pod, Zeroable};
 use egui::plot::{HLine, Line, Plot, Value, Values};
-use egui::{Color32, Ui};
+use egui::{Color32, ColorImage, Ui};
+use egui_vulkano::UpdateTexturesResult;
 use vulkano::buffer::{BufferUsage, CpuAccessibleBuffer, TypedBufferAccess};
-use vulkano::command_buffer::{
-    AutoCommandBufferBuilder, CommandBufferUsage, SubpassContents,
-};
-use vulkano::device::physical::PhysicalDevice;
-use vulkano::device::{Device, DeviceExtensions};
+use vulkano::command_buffer::{AutoCommandBufferBuilder, CommandBufferUsage, SubpassContents};
+use vulkano::device::physical::{PhysicalDevice, PhysicalDeviceType};
+use vulkano::device::{Device, DeviceCreateInfo, DeviceExtensions, QueueCreateInfo};
 use vulkano::format::Format;
 use vulkano::image::view::ImageView;
-use vulkano::image::{ImageUsage, SwapchainImage, ImageAccess};
-use vulkano::instance::Instance;
-use vulkano::pipeline::graphics::viewport::Viewport;
-use vulkano::pipeline::GraphicsPipeline;
-use vulkano::pipeline::graphics::vertex_input::BuffersDefinition;
+use vulkano::image::{ImageAccess, ImageUsage, SwapchainImage};
+use vulkano::instance::{Instance, InstanceCreateInfo};
 use vulkano::pipeline::graphics::input_assembly::InputAssemblyState;
+use vulkano::pipeline::graphics::vertex_input::BuffersDefinition;
+use vulkano::pipeline::graphics::viewport::Viewport;
 use vulkano::pipeline::graphics::viewport::ViewportState;
-use vulkano::render_pass::{Framebuffer, RenderPass, Subpass};
-use vulkano::swapchain::{AcquireError, ColorSpace, Swapchain, SwapchainCreationError};
-use vulkano::sync::{FlushError, GpuFuture};
-use vulkano::{swapchain, sync, Version};
+use vulkano::pipeline::GraphicsPipeline;
+use vulkano::render_pass::{Framebuffer, FramebufferCreateInfo, RenderPass, Subpass};
+use vulkano::swapchain::{AcquireError, Swapchain, SwapchainCreateInfo, SwapchainCreationError};
+use vulkano::sync::{FenceSignalFuture, FlushError, GpuFuture};
+use vulkano::{swapchain, sync};
 use vulkano_win::VkSurfaceBuild;
 use winit::event::{Event, WindowEvent};
 use winit::event_loop::{ControlFlow, EventLoop};
-use winit::window::{Window, WindowBuilder};
+use winit::window::{Fullscreen, Window, WindowBuilder};
+
+pub enum FrameEndFuture<F: GpuFuture + 'static> {
+    FenceSignalFuture(FenceSignalFuture<F>),
+    BoxedFuture(Box<dyn GpuFuture>),
+}
+
+impl<F: GpuFuture> FrameEndFuture<F> {
+    pub fn now(device: Arc<Device>) -> Self {
+        Self::BoxedFuture(sync::now(device).boxed())
+    }
+
+    pub fn get(self) -> Box<dyn GpuFuture> {
+        match self {
+            FrameEndFuture::FenceSignalFuture(f) => f.boxed(),
+            FrameEndFuture::BoxedFuture(f) => f,
+        }
+    }
+}
+
+impl<F: GpuFuture> AsMut<dyn GpuFuture> for FrameEndFuture<F> {
+    fn as_mut(&mut self) -> &mut (dyn GpuFuture + 'static) {
+        match self {
+            FrameEndFuture::FenceSignalFuture(f) => f,
+            FrameEndFuture::BoxedFuture(f) => f,
+        }
+    }
+}
 
 fn main() {
     let required_extensions = vulkano_win::required_extensions();
+    let instance = Instance::new(InstanceCreateInfo {
+        enabled_extensions: required_extensions,
+        ..Default::default()
+    })
+    .unwrap();
 
-    let instance = Instance::new(None, Version::V1_0, &required_extensions, None).unwrap();
     let physical = PhysicalDevice::enumerate(&instance).next().unwrap();
 
     println!(
@@ -48,53 +79,72 @@ fn main() {
     let event_loop = EventLoop::new();
     let surface = WindowBuilder::new()
         .with_title("egui_vulkano demo")
+        .with_fullscreen(Some(Fullscreen::Borderless(None)))
         .build_vk_surface(&event_loop, instance.clone())
         .unwrap();
 
-    let queue_family = physical
-        .queue_families()
-        .find(|&q| q.supports_graphics() && surface.is_supported(q).unwrap_or(false))
-        .unwrap();
-
-    let device_ext = DeviceExtensions {
+    let device_extensions = DeviceExtensions {
         khr_swapchain: true,
         ..DeviceExtensions::none()
     };
+
+    let (physical_device, queue_family) = PhysicalDevice::enumerate(&instance)
+        .filter(|&p| p.supported_extensions().is_superset_of(&device_extensions))
+        .filter_map(|p| {
+            p.queue_families()
+                .find(|&q| q.supports_graphics() && q.supports_surface(&surface).unwrap_or(false))
+                .map(|q| (p, q))
+        })
+        .min_by_key(|(p, _)| match p.properties().device_type {
+            PhysicalDeviceType::DiscreteGpu => 0,
+            PhysicalDeviceType::IntegratedGpu => 1,
+            PhysicalDeviceType::VirtualGpu => 2,
+            PhysicalDeviceType::Cpu => 3,
+            PhysicalDeviceType::Other => 4,
+        })
+        .unwrap();
+
     let (device, mut queues) = Device::new(
-        physical,
-        physical.supported_features(),
-        &physical
-            .required_extensions()
-            .union(&device_ext),
-        [(queue_family, 0.5)].iter().cloned(),
+        physical_device,
+        DeviceCreateInfo {
+            enabled_extensions: physical_device
+                .required_extensions()
+                .union(&device_extensions),
+            queue_create_infos: vec![QueueCreateInfo::family(queue_family)],
+            ..Default::default()
+        },
     )
     .unwrap();
 
     let queue = queues.next().unwrap();
 
     let (mut swapchain, images) = {
-        let caps = surface.capabilities(physical).unwrap();
-        let alpha = caps.supported_composite_alpha.iter().next().unwrap();
+        let caps = physical_device
+            .surface_capabilities(&surface, Default::default())
+            .unwrap();
+        let composite_alpha = caps.supported_composite_alpha.iter().next().unwrap();
 
-        // Set the swapchain format to Srgb to get correct colors for egui
-        assert!(&caps
-            .supported_formats
-            .contains(&(Format::B8G8R8A8_SRGB, ColorSpace::SrgbNonLinear)));
-        let format = Format::B8G8R8A8_SRGB;
-        let dimensions: [u32; 2] = surface.window().inner_size().into();
+        let image_format = Some(Format::B8G8R8A8_SRGB);
+        let image_extent: [u32; 2] = surface.window().inner_size().into();
 
-        Swapchain::start(device.clone(), surface.clone())
-            .num_images(caps.min_image_count)
-            .format(format)
-            .dimensions(dimensions)
-            .usage(ImageUsage::color_attachment())
-            .sharing_mode(&queue)
-            .composite_alpha(alpha)
-            .build()
-            .unwrap()
+        Swapchain::new(
+            device.clone(),
+            surface.clone(),
+            SwapchainCreateInfo {
+                min_image_count: caps.min_image_count,
+                image_format,
+                image_extent,
+                image_usage: ImageUsage::color_attachment(),
+                composite_alpha,
+
+                ..Default::default()
+            },
+        )
+        .unwrap()
     };
 
-    #[derive(Default, Debug, Clone)]
+    #[derive(Default, Debug, Clone, Copy, Pod, Zeroable)]
+    #[repr(C)]
     struct Vertex {
         position: [f32; 2],
     }
@@ -156,21 +206,21 @@ fn main() {
     let fs = fs::load(device.clone()).unwrap();
 
     let render_pass = vulkano::ordered_passes_renderpass!(
-            device.clone(),
-            attachments: {
-                color: {
-                    load: Clear,
-                    store: Store,
-                    format: swapchain.format(),
-                    samples: 1,
-                }
-            },
-            passes: [
-                { color: [color], depth_stencil: {}, input: [] },
-                { color: [color], depth_stencil: {}, input: [] } // Create a second renderpass to draw egui
-            ]
-        )
-        .unwrap();
+        device.clone(),
+        attachments: {
+            color: {
+                load: Clear,
+                store: Store,
+                format: swapchain.image_format(),
+                samples: 1,
+            }
+        },
+        passes: [
+            { color: [color], depth_stencil: {}, input: [] },
+            { color: [color], depth_stencil: {}, input: [] } // Create a second renderpass to draw egui
+        ]
+    )
+    .unwrap();
 
     let pipeline = GraphicsPipeline::start()
         .vertex_input_state(BuffersDefinition::new().vertex::<Vertex>())
@@ -192,12 +242,12 @@ fn main() {
 
     let mut recreate_swapchain = false;
 
-    let mut previous_frame_end = Some(sync::now(device.clone()).boxed());
+    let mut previous_frame_end = Some(FrameEndFuture::now(device.clone()));
 
     //Set up everything need to draw the gui
     let window = surface.window();
-    let mut egui_ctx = egui::CtxRef::default();
-    let mut egui_winit = egui_winit::State::new(window);
+    let egui_ctx = egui::Context::default();
+    let mut egui_winit = egui_winit::State::new(4096, window);
 
     let mut egui_painter = egui_vulkano::Painter::new(
         device.clone(),
@@ -209,9 +259,9 @@ fn main() {
     //Set up some window to look at for the test
 
     let mut egui_test = egui_demo_lib::ColorTest::default();
+    let mut demo_windows = egui_demo_lib::DemoWindows::default();
     let mut egui_bench = Benchmark::new(1000);
-
-    let mut last_frame = Instant::now();
+    let mut my_texture = egui_ctx.load_texture("my_texture", ColorImage::example());
 
     event_loop.run(move |event, _, control_flow| {
         match event {
@@ -234,14 +284,21 @@ fn main() {
                 };
             }
             Event::RedrawEventsCleared => {
-                previous_frame_end.as_mut().unwrap().cleanup_finished();
+                previous_frame_end
+                    .as_mut()
+                    .unwrap()
+                    .as_mut()
+                    .cleanup_finished();
 
                 if recreate_swapchain {
                     let dimensions: [u32; 2] = surface.window().inner_size().into();
                     let (new_swapchain, new_images) =
-                        match swapchain.recreate().dimensions(dimensions).build() {
+                        match swapchain.recreate(SwapchainCreateInfo {
+                            image_extent: surface.window().inner_size().into(),
+                            ..swapchain.create_info()
+                        }) {
                             Ok(r) => r,
-                            Err(SwapchainCreationError::UnsupportedDimensions) => return,
+                            Err(SwapchainCreationError::ImageExtentNotSupported { .. }) => return,
                             Err(e) => panic!("Failed to recreate swapchain: {:?}", e),
                         };
 
@@ -277,35 +334,14 @@ fn main() {
                 )
                 .unwrap();
 
-                // Do your usual rendering
-                builder
-                    .begin_render_pass(
-                        framebuffers[image_num].clone(),
-                        SubpassContents::Inline,
-                        clear_values,
-                    )
-                    .unwrap()
-                    .set_viewport(0, [viewport.clone()])
-                    .bind_pipeline_graphics(pipeline.clone())
-                    .bind_vertex_buffers(0, vertex_buffer.clone())
-                    .draw(
-                        vertex_buffer.len().try_into().unwrap(),
-                        1,
-                        0,
-                        0,
-                    )
-                    .unwrap(); // Don't end the render pass yet
-
-                // Build your gui
-                egui_bench.push(last_frame.elapsed().as_secs_f64());
-                last_frame = Instant::now();
-
+                let frame_start = Instant::now();
                 egui_ctx.begin_frame(egui_winit.take_egui_input(surface.window()));
+                demo_windows.ui(&egui_ctx);
 
                 egui::Window::new("Color test")
                     .vscroll(true)
                     .show(&egui_ctx, |ui| {
-                        egui_test.ui(ui, None);
+                        egui_test.ui(ui);
                     });
 
                 egui::Window::new("Settings").show(&egui_ctx, |ui| {
@@ -318,9 +354,40 @@ fn main() {
                         egui_bench.draw(ui);
                     });
 
+                egui::Window::new("Texture test").show(&egui_ctx, |ui| {
+                    ui.image(my_texture.id(), (200.0, 200.0));
+                    if ui.button("Reload texture").clicked() {
+                        // previous TextureHandle is dropped, causing egui to free the texture:
+                        my_texture = egui_ctx.load_texture("my_texture", ColorImage::example());
+                    }
+                });
+
                 // Get the shapes from egui
-                let (egui_output, clipped_shapes) = egui_ctx.end_frame();
-                egui_winit.handle_output(surface.window(), &egui_ctx, egui_output);
+                let egui_output = egui_ctx.end_frame();
+                let platform_output = egui_output.platform_output;
+                egui_winit.handle_platform_output(surface.window(), &egui_ctx, platform_output);
+
+                let result = egui_painter
+                    .update_textures(egui_output.textures_delta, &mut builder)
+                    .expect("egui texture error");
+
+                let wait_for_last_frame = result == UpdateTexturesResult::Changed;
+
+                // Do your usual rendering
+                builder
+                    .begin_render_pass(
+                        framebuffers[image_num].clone(),
+                        SubpassContents::Inline,
+                        clear_values,
+                    )
+                    .unwrap()
+                    .set_viewport(0, [viewport.clone()])
+                    .bind_pipeline_graphics(pipeline.clone())
+                    .bind_vertex_buffers(0, vertex_buffer.clone())
+                    .draw(vertex_buffer.len().try_into().unwrap(), 1, 0, 0)
+                    .unwrap(); // Don't end the render pass yet
+
+                // Build your gui
 
                 // Automatically start the next render subpass and draw the gui
                 let size = surface.window().inner_size();
@@ -328,20 +395,29 @@ fn main() {
                 egui_painter
                     .draw(
                         &mut builder,
-                        [(size.width as f32)/sf, (size.height as f32)/sf],
+                        [(size.width as f32) / sf, (size.height as f32) / sf],
                         &egui_ctx,
-                        clipped_shapes,
+                        egui_output.shapes,
                     )
                     .unwrap();
 
+                egui_bench.push(frame_start.elapsed().as_secs_f64());
+
                 // End the render pass as usual
                 builder.end_render_pass().unwrap();
 
                 let command_buffer = builder.build().unwrap();
 
+                if wait_for_last_frame {
+                    if let Some(FrameEndFuture::FenceSignalFuture(ref mut f)) = previous_frame_end {
+                        f.wait(None).unwrap();
+                    }
+                }
+
                 let future = previous_frame_end
                     .take()
                     .unwrap()
+                    .get()
                     .join(acquire_future)
                     .then_execute(queue.clone(), command_buffer)
                     .unwrap()
@@ -350,15 +426,15 @@ fn main() {
 
                 match future {
                     Ok(future) => {
-                        previous_frame_end = Some(future.boxed());
+                        previous_frame_end = Some(FrameEndFuture::FenceSignalFuture(future));
                     }
                     Err(FlushError::OutOfDate) => {
                         recreate_swapchain = true;
-                        previous_frame_end = Some(sync::now(device.clone()).boxed());
+                        previous_frame_end = Some(FrameEndFuture::now(device.clone()));
                     }
                     Err(e) => {
                         println!("Failed to flush future: {:?}", e);
-                        previous_frame_end = Some(sync::now(device.clone()).boxed());
+                        previous_frame_end = Some(FrameEndFuture::now(device.clone()));
                     }
                 }
             }
@@ -375,18 +451,20 @@ fn window_size_dependent_setup(
     let dimensions = images[0].dimensions().width_height();
     viewport.dimensions = [dimensions[0] as f32, dimensions[1] as f32];
 
-    let framebuffers = images
+    images
         .iter()
         .map(|image| {
-            let view = ImageView::new(image.clone()).unwrap();
-            Framebuffer::start(render_pass.clone())
-                .add(view)
-                .unwrap()
-                .build()
-                .unwrap()
+            let view = ImageView::new_default(image.clone()).unwrap();
+            Framebuffer::new(
+                render_pass.clone(),
+                FramebufferCreateInfo {
+                    attachments: vec![view],
+                    ..Default::default()
+                },
+            )
+            .unwrap()
         })
-        .collect::<Vec<_>>();
-    framebuffers
+        .collect::<Vec<_>>()
 }
 
 pub struct Benchmark {
@@ -409,21 +487,17 @@ impl Benchmark {
             .enumerate()
             .map(|(i, v)| Value::new(i as f64, *v * 1000.0));
         let curve = Line::new(Values::from_values_iter(iter)).color(Color32::BLUE);
-        // As of egui 0.13.0 it is impossible to set the color of an HLine
-        // see https://github.com/emilk/egui/issues/524
-        // let red = Stroke::new(2.0, Color32::RED);
-        let zero = HLine::new(0.0);
-        let target = HLine::new(1000.0 / 60.0);
+        let target = HLine::new(1000.0 / 60.0).color(Color32::RED);
 
-        ui.label("Time in milliseconds that each frame took to draw:");
+        ui.label("Time in milliseconds that the gui took to draw:");
         Plot::new("plot")
             .view_aspect(2.0)
+            .include_y(0)
             .show(ui, |plot_ui| {
                 plot_ui.line(curve);
-                plot_ui.hline(zero);
-                plot_ui.hline(target);
+                plot_ui.hline(target)
             });
-        ui.label("The light blue line marks the frametime target for drawing at 60 FPS.");
+        ui.label("The red line marks the frametime target for drawing at 60 FPS.");
     }
 
     pub fn push(&mut self, v: f64) {
